/*
 * $Id: main.cpp,v 1.3 2004/05/14 09:41:11 tsuruoka Exp $
 */

#include <stdio.h>
#include <string>
#include <iostream>
#include <fstream>
#include <list>
#include <vector>
#include "pndist.h"
#include <map>
#include <algorithm>
#include <set>
#include "atmstr.h"
#include "trie.h"

using namespace std;

const int MAX_N = 200;

int npairs = 0;

bool interactive_mode = true;
//const bool interactive_mode = false;

Trie chunk_dic;

string::size_type
find_best_long_form(const string& shortform, const string& longform);

set<string> short_long;
map<string, string> short2long;
map<string, string> long2short;


struct GString {
  enum { UNINITIALIZED, IS_CHAR, IS_INT };
  GString() {
    _mode = UNINITIALIZED;
  }
  GString(const string & s) {
    //    _mode = IS_CHAR;
    _mode = IS_INT;
    _v.resize(s.size());
    for (int i = 0; i < s.size(); i++) _v[i] = s[i];
  }
  GString(const vector<int> & s) {
    _mode = IS_INT;
    _v = s;
    //    _v.resize(s.size());
    //    for (int i = 0; i < s.size(); i++) _v[i] = s[i];
  }
  string str() const {
    string s;
    switch (_mode) {
    case UNINITIALIZED:
      cerr << "error: GString::str()" << endl;
      exit(1);
    case IS_CHAR:
      for (int i = 0; i < _v.size(); i++) s.push_back(char(_v[i]));
      break;
    case IS_INT:
      for (int i = 0; i < _v.size(); i++) {
        //        if (i != 0) s += " ";
        s += AString(_v[i]).str();
      }
      break;
    }
    return s;
  }
  int size() const { return _v.size(); }
  int & operator[](const int index) { return _v[index]; }
  const int & operator[](const int index) const { return _v[index]; }
  vector<int> substr(const int begin, const int num) const {
    assert(begin + num <= _v.size());
    vector<int> tmp(&_v[begin], &_v[begin + num]);
    return tmp;
  }
  bool operator<(const GString & x) const {
    return this->_v < x._v;
  }
  bool operator==(const GString & x) const {
    return this->_v == x._v;
  }
private:
  int _mode;
  vector<int> _v;
  //  basic_string<int> _v;
};

//typedef basic_string<int> GString;

struct Rule {
  GString context;
  Operation op;
  double score;
  Rule(const GString &s, const Operation &o, double sc) {
    context = s;
    op = o;
    score = sc;
  }
  bool operator<(const Rule &r) {
    return (score > r.score);
  }
  void Print() const {
    printf("%5.3f %s ", score, context.str().c_str());
    op.print();
  }
};

class RuleBag
{
  struct Elem {
    list<Operation> lop;
    int n;
    Elem(): n(0) {};
  };
  map<GString, Elem> me;

  double calc_score(int k, int n) {
    return (k + 1.0) / (n + 2.0);
    //    return (k + 0.1) / (n + 1.0);
  }
public:
  void Put(const GString& c) {
    me[c].n += 1;
  }
  void Put(const GString& c, const Operation& op) {
    Elem& e = me[c];
    //    e.n += 1;
    //    list<Operation>::iterator i = e.lop.find(op);
    list<Operation>::iterator i = find(e.lop.begin(), e.lop.end(), op);
    if (i != e.lop.end()) {
      i->k += 1;
    } else {
      Operation o(op);
      o.k += 1;
      e.lop.push_back(o);
    }
  }
  bool Rules(const GString &c, list<Rule> &lr) {
    map<GString, Elem>::const_iterator i = me.find(c);
    if (i == me.end()) return false;
    if (i->second.lop.empty()) return false;
    for (list<Operation>::const_iterator j = i->second.lop.begin();
         j != i->second.lop.end(); j++) {
      double score = calc_score(j->k, i->second.n);
      Rule r(i->first, *j, score);
      lr.push_back(r);
    }
    return true;
  }
  void Print() {
    for (map<GString, Elem>::const_iterator i = me.begin(); i != me.end(); i++) {
      if (i->second.lop.empty()) continue;
      cout << "context: " << i->first.str() << " [" << i->second.n << "]" << endl;
      for (list<Operation>::const_iterator j = i->second.lop.begin();
           j != i->second.lop.end(); j++) {
        cout << "operation: ";
        j->print();
        cout << " [" << j->k << "]" << endl;
      }
      cout << endl;
    }
  }
  void OutputRules(list<Rule> &lr) {
    for (map<GString, Elem>::const_iterator i = me.begin(); i != me.end(); i++) {
      if (i->second.n == 0) continue;
      for (list<Operation>::const_iterator j = i->second.lop.begin();
           j != i->second.lop.end(); j++) {
        double score = calc_score(j->k, i->second.n);
        Rule r(i->first, *j, score);
        lr.push_back(r);
      }
    }
  }
  
};

RuleBag rb;



void read_term_list(const string& filename, vector<string>& vs)
{
  ifstream ifile(filename.c_str());
  if (!ifile) {
    cerr << "cannot open " << filename << "!" << endl;
    exit(1);
  }
  //  cerr << "Reading " << filename;

  string line;
  while(getline(ifile, line)) {
    vs.push_back(line);
  }
  //  cerr << endl;

  ifile.close();
}

void expand_context(const GString &s1, list<GString> &c)
{
  // abcdefg
  GString s;
  //  c.push_back(s1);

  // *bcdef*
  s = s1;
  s[0] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);

  // **cde**
  s = s1;
  s[0] = '*';
  s[1] = '*';
  s[s.size()-2] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);

  // ***d***
  s = s1;
  s[0] = '*';
  s[1] = '*';
  s[2] = '*';
  s[s.size()-3] = '*';
  s[s.size()-2] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);

  // ***de**
  s = s1;
  s[0] = '*';
  s[1] = '*';
  s[2] = '*';
  s[s.size()-2] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);

  // **cd***
  s = s1;
  s[0] = '*';
  s[1] = '*';
  s[s.size()-3] = '*';
  s[s.size()-2] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);

  // *bcd***
  s = s1;
  s[0] = '*';
  s[s.size()-3] = '*';
  s[s.size()-2] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);

  // ***def*
  s = s1;
  s[0] = '*';
  s[1] = '*';
  s[2] = '*';
  s[s.size()-1] = '*';
  c.push_back(s);
}

string apply_rule(const string &s1, const Operation &op, int pos)
{
  string s = "{{{" + s1 + "}}}";
  /*
  string s = "{{{" + string(s1.size(), '#') + "}}}";
  for (int i = 0; i < s1.size(); i++) {
    s[i+3] = s1[i];
  }
  */
  pos += 3;

  switch (op.type) {
  case Operation::NONE: {
    return s1;
  }
  case Operation::SUB: {
    //    s.replace(pos, 1, 1, op.to);
    string from_str = AString(op.from).str();
    string to_str   = AString(op.to).str();
    s.replace(pos, from_str.size(), to_str);
    break;
  }
  case Operation::DEL: {
    //    s.replace(pos, 1, 0, ' ');
    int len = AString(op.chr).str().size();
    s.replace(pos, len, "");
    break;
  }
  case Operation::INS: {
    //    s.replace(pos, 0, 1, op.chr);
    s.replace(pos, 0, AString(op.chr).str());
    break;
  }
  }

  return s.substr(3, s.size()-6);
}

struct Candidate {
  string str;
  double score;
  Candidate(const string &s, double sc) : str(s), score(sc) {};
  bool operator==(const Candidate &c) const { return str == c.str; }
};
  
void expand_string(const string &s1)
{
  list<Candidate> ls;

  ls.push_back(Candidate(s1, 1.0));

  if (interactive_mode) {
    printf("%5.3f %s\n", 1.0, s1.c_str());
  } else {
    //    cout << s1 << endl;
    printf("%5.3f %s\n", 1.0, s1.c_str());
    //    if (s1.size() <= 5) return;
  }

  while(ls.size() < 10) {
    double best_score = 0;
    string best_str;

    Operation dummy;
    Rule best_rule(string("dummy"), dummy, 0);
    for (list<Candidate>::const_iterator ii = ls.begin(); ii != ls.end(); ii++) {
      if (ii->score <= best_score) continue;
    
    //  cout << s1 << endl;
      string s1 = ii->str;
      string s = "{{{" + s1 + "}}}";
      // SUB, DEL
      for (int i = 0; i < s1.size(); i++) {
        GString c = s.substr(i, 7);
        list<GString> lct;
        expand_context(c, lct);

        // for tokens
        int contain = chunk_dic.LongestMatch(s1, i);
        if (contain) {
          for (int end = i + 2; end <= s1.size(); end++) {
            string ss = s1.substr(i, end - i);
            if (!chunk_dic.Lookup(ss)) continue;
            GString c = s.substr(i, 3) + "#" + s.substr(end + 3, 3);
            c[3] = AString(ss);
            expand_context(c, lct);
          }
        }
        
        for (list<GString>::const_iterator j = lct.begin(); j != lct.end(); j++) {
          list<Rule> lr;
          rb.Rules(*j, lr);
          for (list<Rule>::const_iterator k = lr.begin(); k != lr.end(); k++) {
            double score = ii->score * k->score;
            //        k->Print();
            //        cout << "\t" << apply_rule(s1, k->op, i);
            //        cout << endl;
            if (score > best_score) {
              string a = apply_rule(s1, k->op, i);
              if (find(ls.begin(), ls.end(), Candidate(a,0)) == ls.end()) {
                best_rule = *k;
                best_score = score;
                best_str = a;
              }
            }
          }
        }
      }

      // INS
      for (int i = 0; i <= s1.size(); i++) {
        string c = s.substr(i, 3) + "_" + s.substr(i+3, 3);
        list<GString> lct;
        expand_context(c, lct);
        for (list<GString>::const_iterator j = lct.begin(); j != lct.end(); j++) {
          list<Rule> lr;
          rb.Rules(*j, lr);
          for (list<Rule>::const_iterator k = lr.begin(); k != lr.end(); k++) {
            double score = ii->score * k->score;
            //          k->Print();
            //          cout << "\t" << apply_rule(s1, k->op, i);
            //          cout << endl;
            if (score > best_score) {
              string a = apply_rule(s1, k->op, i);
              if (find(ls.begin(), ls.end(), Candidate(a,0)) == ls.end()) {
                best_rule = *k;
                best_score = score;
                best_str = a;
              }
            }
          }
        }
      }
    }
    if (interactive_mode) {
      if (best_str.size() < 40)
        printf("%5.3f %-40s  ", best_score, best_str.c_str());
      else
        printf("%5.3f %-60s  ", best_score, best_str.c_str());
      //      best_rule.Print();
      cout << endl;
    } else {
      if (best_score < 0.007) break;
      printf("%5.3f %s\n", best_score, best_str.c_str());
      //      cout << best_str.c_str() << endl;
    }
    ls.push_back(Candidate(best_str, best_score));
  }

}

void extract_context(const string& s1, list<GString>& contexts)
{
  string s = "{{{" + s1 + "}}}";

  for (int i = 0; i < s1.size(); i++) {
    GString c = s.substr(i, 7);
    expand_context(c, contexts);
  }

  for (int i = 0; i <= s1.size(); i++) {
    GString c = s.substr(i, 3) + "_" + s.substr(i+3, 3);
    expand_context(c, contexts);
  }

}

void extract_chunk_contexts(const string & s,
                            list<GString> & contexts)
{
  string s1 = "{{{" + s + "}}}";
  
  for (int i = 0; i < s.size(); i++) {
    int contain = chunk_dic.LongestMatch(s, i);
    if (contain) {
      for (int end = i + 2; end <= s.size(); end++) {
        string ss = s.substr(i, end - i);
        if (!chunk_dic.Lookup(ss)) continue;
        GString c = s1.substr(i, 3) + "#" + s1.substr(end + 3, 3);
        c[3] = AString(ss);
        expand_context(c, contexts);
        //        GString gs;
        //        expand_context();
        // the contexts for insertion are generated in extract_context()
      }
    }
  }



}
                            

void extract_contexts(vector<string> & term_vec)
{
  cerr << "extracting contexts";
  for (int i = 0; i < term_vec.size(); i++) {

    // put possible contexts into the rulebag
    const string & s = term_vec[i];
    list<GString> contexts;
    extract_context(s, contexts);
    extract_chunk_contexts(s, contexts);
    for (list<GString>::const_iterator k = contexts.begin();
         k != contexts.end(); k++) {
      rb.Put(*k);
    }
    if (i % 100 == 0) cerr << ".";
  }
  cerr << endl;
}



bool
find_common_chunk(const string & s1, const string & s2,
                int *b1, int *e1,
                int *b2, int *e2)
{
  const static string separators = " -,./";
  int left = 0, sep_left = 0;
  for (int i = 0; i < s1.size(); i++) {
    if (s1[i] != s2[i]) break;
    left++;
    if (separators.find(s1[i]) != string::npos) sep_left = left;
  }
  int right = 0, sep_right = 0;
  for (int i = 0; i < s1.size(); i++) {
    int i1 = s1.size() - 1 - i;
    if (s1[i1] != s2[s2.size() - 1 - i]) break;
    if (left + right >= s1.size()) break;
    if (left + right >= s2.size()) break;
    right++;
    if (separators.find(s1[i1]) != string::npos) sep_right = right;
  }
  left = sep_left;
  right = sep_right;

  *b1 = left;
  *e1 = s1.size() - right;
  *b2 = left;
  *e2 = s2.size() - right;

  return true;

  string chunk1 = s1.substr(left, s1.size() - right - left);
  string chunk2 = s2.substr(left, s2.size() - right - left);
  cout << "s1 = " << s1 << endl;
  cout << "s2 = " << s2 << endl;
  cout << "left = " << left << " right = " << right << endl;
  cout << "chunk1 = " << chunk1 << endl;
  cout << "chunk2 = " << chunk2 << endl;
  cout << endl;

  return true;
  
  //  if (left < 1 && right < 1)  return false;


  
}

vector<int>
string2intvec(const string & s)
{
  vector<int> v(s.size());
  for (int i = 0; i < s.size(); i++) {
    v[i] = s[i];
  }
  return v;
}


vector<int>
tokenize(const string & s)
{
  vector<int> v;

  int start = 0;
  for (int i = 0; ; i++) {
    if (i == s.size() || s[i] == ' ') {
      string token = s.substr(start, i - start);
      if (token.size() > 0) {
        v.push_back(AString(token));
        //        cout << token << endl;
      }
      if (i == s.size()) break;
      start = i + 1;
    }
  }
  return v;
}

int extract_operations(const vector<string>& term_vec)
{
  //  if (term_vec.size() <= 1) {
  //    return -1;
  //  }
  if (term_vec.size() >= MAX_N) {
    cerr << "error: term_vec.size() >= MAX_N" << endl;
    exit(1);
  }
  
  //  for (int i = 0; i < term_vec.size(); i++)
  //    cout << i << "\t" << term_vec[i] << endl;
  //  cout << endl;

  int n = term_vec.size();

  //  if (n == 1) cout << term_vec[0] << endl;
  //  return 0;
                
  if (n > MAX_N) n = MAX_N;
  for (int i = 0; i < n; i++) {

    for (int j = 0; j < n; j++) {
      if (i == j) continue;
      const string & s1 = term_vec[i];
      const string & s2 = term_vec[j];
      int b, e, pos;
      Operation op;
      int c = editdistance(&string2intvec(s1)[0], s1.size(),
                           &string2intvec(s2)[0], s2.size(),
                           &b, &e, op, pos);
      if (c == 1) {
        npairs++;
        string s = "{{{" + s1 + "}}}";
        string ct;
        switch (op.type) {
        case Operation::NONE:
          break;
        case Operation::SUB:
          ct = s.substr(pos-1, 7);
          break;
        case Operation::DEL:
          ct = s.substr(pos-1, 7);
          break;
        case Operation::INS: 
          ct = s.substr(pos, 3) + "_" + s.substr(pos+3,3);
          break;
        }
        if (op.type == Operation::NONE) {
          cerr << "warning: op.type == NONE" << endl;
          cerr << s1 << endl;
          cerr << s2 << endl;
          continue;
        }
        list<GString> ls;
        expand_context(ct, ls);
        for (list<GString>::const_iterator k = ls.begin(); k != ls.end(); k++) {
          rb.Put(*k, op);
        }
        continue;
      }
      // if cost != 1
      int b1, e1, b2, e2;
      find_common_chunk(s1, s2, &b1, &e1, &b2, &e2);
      assert(e1 >= b1 && e2 >= b2);
      if (e1 - b1 > 0 || e2 - b2 > 0) {
        const string chunk1 = s1.substr(b1, e1 - b1);
        const string chunk2 = s2.substr(b2, e2 - b2);
        chunk_dic.Insert(chunk1);
        chunk_dic.Insert(chunk2);
        string s = "{{{" + s1 + "}}}";
        GString ct;
        if (e1 - b1 == 0) {
          // insertion
          ct = s.substr(b1, 3) + "_" + s.substr(e1 + 3,3);
          op.type = Operation::INS;
          op.chr = AString(chunk2);
        } else if (e2 - b2 == 0) {
          // deletion
          ct = s.substr(b1, 3) + "#" + s.substr(e1 + 3, 3);
          ct[3] = AString(chunk1);
          op.type = Operation::DEL;
          op.chr = AString(chunk1);
        } else {
          // substitution
          ct = s.substr(b1, 3) + "#" + s.substr(e1 + 3, 3);
          static int n = 0;
          n++;
          ct[3] = AString(chunk1);
          assert(ct[3] >= 0);
          op.type = Operation::SUB;
          op.from = AString(chunk1);
          op.to = AString(chunk2);
        }
        list<GString> ls;
        expand_context(ct, ls);
        for (list<GString>::const_iterator k = ls.begin(); k != ls.end(); k++) {
          rb.Put(*k, op);
        }

      }

      
    }
  }


}

int main(int argc, char** argv)
{
  if (argc == 2 && string(argv[1]) == "--batch") {
    interactive_mode = false;
  }
  // to avoid index conflict between characters and tokens
  for (int i = 0; i < 256; i++) {
    char buf[2];
    buf[0] = i; buf[1] = 0;
    AString a(buf);
  }
  if (AString("test") != 256) {
    cerr << "initialization error" << endl;
    exit(1);
  }
  
  cerr << "---------------------------------------------------------" << endl;
  cerr << "   Probabilistic Term Variant Generator version 0.3" << endl;
  cerr << "---------------------------------------------------------" << endl;
  cerr << "extracting operations";
      
  vector<string> term_vec, tmp_term_vec;

  //  read_term_list("termlist.txt", term_vec);
  //  ifstream ifile("human_cell2000-2001.acronyms");
  //  ifstream ifile("umls_protein.clusters");
  ifstream ifile("train.txt");
  string line;
  int n = 0;
  int nshort = 0;
  int nlong = 0;
  while (getline(ifile, line)) {
#ifndef NDEBUG
    if (nshort > 1000) {
      cerr << endl << "!!!<debug mode> using only 1000 clusters!!!";
      break;
    }
#endif      
    if (line.size() == 0) {
      extract_operations(tmp_term_vec);
      getline(ifile, line);  // read short form;
      nshort++;
      tmp_term_vec.clear();
      continue;
    }
    tmp_term_vec.push_back(line);
    term_vec.push_back(line);
    nlong++;
    if (++n % 100 == 0) cerr << ".";
  }
  cerr << endl;
  extract_contexts(term_vec);
  
  cerr << endl;
  cerr << "nshort = " << nshort << " nlong = " << nlong << endl;
  cerr << "nparis = " << npairs << endl;
  ifile.close();
  //  rb.Print();
  list<Rule> lr;
  rb.OutputRules(lr);
  lr.sort();

  if (interactive_mode)
    cout << "///// Operation Probability" << endl;

  n = 0;
  if (interactive_mode) {
    cout << "nrules = " << lr.size() << endl;
    for (list<Rule>::const_iterator i = lr.begin(); i != lr.end(); i++) {
      i->Print();
      cout << endl;
      if (++n > 100) break;
    }
    cout << "  :" << endl;
  }

  if (interactive_mode)
    //    cout << endl << "Input a biomedical term (e.g. IkappaBalpha, anemia, activated T cell, ...)" << endl;
    cout << endl << "Input a protein name." << endl;
  list<string> ls;
  while (1) {
    if (interactive_mode)
      cout << "> ";
    if (!getline(cin, line)) break;
    //    if (line == "q") break;
    expand_string(line);
    //    cout << endl;
  }
  //  expand_string("NF-kappa B");
  //  expand_string("multipurpose");
  //  expand_string("anaemia");
  
}

/*
 * $Log: main.cpp,v $
 * Revision 1.3  2004/05/14 09:41:11  tsuruoka
 * add find_chunk_diff()
 *
 * Revision 1.2  2004/05/14 05:56:39  tsuruoka
 * add GString
 *
 * Revision 1.1  2004/05/14 04:40:00  tsuruoka
 * rename cluster.cpp -> main.cpp
 *
 * Revision 1.4  2004/05/13 13:29:46  tsuruoka
 * enable editdistance() to handle vector<int>
 *
 */
