// Editdistance for protein names

#include <ctype.h>
#include <iostream>
#include <assert.h>
#include <limits.h>
#include "pndist.h"

int  DPTable::tabp[(DPTable::MAX_N + 1)*(DPTable::MAX_M + 1)];

int editdistance(const int* ar1, int n,
                 const int* ar2, int m,
                 int* begin, int* end,
                 Operation& op, int& op_pos)
{
  
  DPTable tbl(ar1, n, ar2, m);

  int p1, p2, p3;
  int nPrice;
  int maxnm = (n>m ? n : m);
  int minnm = std::min(n, m);

  int m_xmax = n;
  int m_ymax = m;

  // Fill the first row and column of graph
  tbl.Set(0, 0, 0);
  for (int x = 1; x <= n; x++) {
    tbl.Set(x, 0, tbl.Get(x-1,0) + tbl.InsertCost(ar1[x-1],x,0));
  }
  
  for (int y = 1; y <= m; y++)
    tbl.Set(0, y, tbl.Get(0, y-1) + tbl.DeleteCost(ar2[y-1],0,y));

  // Fill the rest of the graph
  for (int y = 1; y <= m; y++) {
    int min_cost = INT_MAX/4;
    for (int x = 1; x <= n; x++) {
      int cost = tbl.Min(x, y);
      assert(cost >= 0);
      tbl.Set(x, y, cost);
      if (cost < min_cost) min_cost = cost;
    }
    //    if (min_cost > 1) return min_cost;
  }


  int min = tbl.Get(n, m);
  int pos = n;
  
  // Traverse the table to find the beginning letter
  int x = pos;
  int y = m;
  while(1) {
    int c1 = INT_MAX;
    if (x > 0 && y > 0)
      c1 = tbl.Get(x-1, y-1) + tbl.ChangeCost(ar1[x-1], ar2[y-1], x, y);
    int c2 = INT_MAX;
    if (y > 0) 
      c2 = tbl.Get( x , y-1) + tbl.InsertCost(ar2[y-1],x,y);
    int c3 = INT_MAX;
    if (x > 0)
      c3 = tbl.Get(x-1,  y)  + tbl.DeleteCost(ar1[x-1],x,y);
    int c = tbl.Get(x, y);
    if (c1 <= c2) {
      if (c1 <= c3) {
        if (c - tbl.Get(x-1,y-1) == 1) {
          op.type = Operation::SUB;
          op.from = ar1[x-1];
          op.to   = ar2[y-1];
          op_pos = x;
        }
        x--; y--;  // substitution
      }
      else         {
        if (c - tbl.Get(x-1,y) == 1) {
          op.type = Operation::DEL;
          op.chr  = ar1[x-1];
          op_pos = x;
        }
        x--;
      }
    } else {
      if (c2 < c3) {
        if (c - tbl.Get(x, y-1) == 1) {
          op.type = Operation::INS;
          op.chr  = ar2[y-1];
          op_pos = x;
        }
        y--;
      }
      else         {
        if (c - tbl.Get(x-1,y) == 1) {
          op.type = Operation::DEL;
          op.chr  = ar1[x-1];
          op_pos = x;
        }
        x--;
      }
    }
    if (y == 0 && x == 0) break;
  }
  *begin = x;
  *end = pos;

  assert(min >= 0);
  return min;
}
