#ifndef PNDIST_H
#define PNDIST_H

#include <stdlib.h>
#include <algorithm>
#include <string>
#include <iostream>
#include "atmstr.h"

struct Operation {
  enum operation_type { NONE, DEL, INS, SUB };
  operation_type type;
  //  string lcon;
  //  string rcon;
  int from;
  int to;
  int chr;
  int k;
  Operation() : type(NONE), k(0), from('#'), to('#'), chr('#') {};
  bool operator==(const Operation& o) const {
    if (type != o.type) return false;
    if (from != o.from) return false;
    if (to   != o.to)   return false;
    if (chr  != o.chr)  return false;
    return true;  // don't care about the value of 'k'
  }
  void print() const {
    switch (this->type) {
    case Operation::NONE: {
      std::cout << "NONE";
      break;
    }
    case Operation::SUB: {
      std::cout << "SUB " << AString(this->from).str() << " " << AString(this->to).str();
      break;
    }
    case Operation::DEL: {
      std::cout << "DEL " << AString(this->chr).str();
      break;
    }
    case Operation::INS: {
      std::cout << "INS " << AString(this->chr).str();
      break;
    }
    }

  }

};

int editdistance(const int* ar1, int n,
                 const int* ar2, int m,
                 int* begin, int* end,
                 Operation&, int& pos);

class DPTable
{
public:
  enum tablesize {
    MAX_N = 2000,
    MAX_M = 2000,
  };
  int DeleteCost(const int deleted, int x, int y) {
    return 1;
  }
  int InsertCost(const int inserted, int x, int y) {
    return 1;
  }
  int ChangeCost(const int from, const int to, int x, int y) {
    if (from == to) return 0;
    return 1;
  }
public:
  DPTable(const int* ar1_, int n_, const int* ar2_, int m_) {
    if (n_ > MAX_N) {
      std::cerr << "error: n > MAX_N" << std::endl;
      exit(1);
    }
    if (m_ > MAX_M) {
      std::cerr << "error: m > MAX_M" << std::endl;
      exit(1);
    }
    n = n_; m = m_;
    ar1 = ar1_; ar2 = ar2_;
  }
  ~DPTable() {
    //    delete[] tabp;
  }
  void Set(int x, int y, int val) {
    assert(x >= 0 && y >= 0);
    assert(x <= n && y <= m);
    //    tabp[(n+1)*y + x] = val;
    tabp[(MAX_N+1)*y + x] = val;
  }
  int Get(int x, int y) {
    assert(x >= 0 && y >= 0);
    assert(x <= n && y <= m);
    //    return tabp[(n+1)*y + x];
    return tabp[(MAX_N+1)*y + x];
  }
  int Min(int x, int y) {
    assert(x >= 1 && y >= 1);
    assert(x <= n && y <= m);
    int p1 = tabp[x-1 +(MAX_N+1)*(y-1)] + ChangeCost(ar1[x-1], ar2[y-1],x,y);
    int p2 = tabp[ x  +(MAX_N+1)*(y-1)] + InsertCost(ar2[y-1],x,y);
    int p3 = tabp[x-1 +(MAX_N+1)*  y  ] + DeleteCost(ar1[x-1],x,y);
    int cost = std::min(std::min(p1,p2),p3);
    return cost;
  }
private:
  static int tabp[(MAX_N+1)*(MAX_M+1)];
  const int *ar1, *ar2;
  int n, m;   // string length
};


#endif
