#include <algorithm>
#include <list>

class Trie
{
  struct Node;
  typedef std::pair<char,Node*> Edge;
  struct Node {
    std::list<Edge> edges;
  };
public:
  Trie() {
    root = new Node;
  };
  ~Trie() {};
  bool Insert(const std::string& s) {
    return _insert(s, 0, root);
  };
  bool Delete(const std::string& s);
  bool Lookup(const std::string& s) const {
    return _lookup(s, 0, root);
  };
  int LongestMatch(const std::string &s, int start_index = 0,
                   const char *delimiters = NULL) const {
    return _longest(s, start_index, root, delimiters);
  };
private:
  class first_equals {
    char value;
  public:
    first_equals(const char k) : value(k) {}
    bool operator() (std::pair<char, Node*> elem) {
      return elem.first == value;
    }
  };
  typedef std::list<Edge>::const_iterator LACI;

  Node* root;
  bool _insert(const std::string& s, unsigned int idx, Node* np) {
    assert(idx <= s.size());
    if (idx == s.size()) {
      if (find_if(np->edges.begin(), np->edges.end(), first_equals(0))
          == np->edges.end())
        np->edges.push_back(std::make_pair(0, (Node*)NULL));
      return true;
    }
    LACI i = find_if(np->edges.begin(), np->edges.end(), first_equals(s[idx]));
    if (i != np->edges.end())
      return _insert(s, idx + 1, i->second);
            
    Node* nnp = new Node;
    np->edges.push_back(std::make_pair(s[idx], nnp));
    return _insert(s, idx + 1, nnp);
  };
  bool _lookup(const std::string& s, unsigned int idx, Node* np) const {
    assert(idx <= s.size());
    if (idx == s.size()) {
      LACI i = find_if(np->edges.begin(), np->edges.end(), first_equals(0));
      return (i != np->edges.end());
    }
    LACI i = find_if(np->edges.begin(), np->edges.end(), first_equals(s[idx]));
    if (i != np->edges.end())
      return _lookup(s, idx + 1, i->second);
    return false;
  };
  int _longest(const std::string& s, unsigned int idx, Node* np,
               const char *delimiters) const {
    assert(idx <= s.size());
    if (idx == s.size()) {
      LACI i = find_if(np->edges.begin(), np->edges.end(), first_equals(0));
      if (i == np->edges.end()) return 0;
      else                      return idx;
    }
    LACI i = find_if(np->edges.begin(), np->edges.end(), first_equals(s[idx]));
    if (i != np->edges.end()) {
      int longer = _longest(s, idx + 1, i->second, delimiters);
      if (longer != 0) return longer;
    }
    i = find_if(np->edges.begin(), np->edges.end(), first_equals(0));
    if (i == np->edges.end()) return 0;
    else {
      if (delimiters) {
        int n = 0, found = 0;
        char c;
        while ( (c = delimiters[n]) != 0) {
          if (c == s[idx]) { found = 1; break; }
          n++;
        }
        if (found) return idx;
        else       return 0;
      } else {
        return idx;
      }
    }
  };
  
};
